# trimgmi
# Copyright 2022 David Seaward and contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import string
from dataclasses import dataclass, field
from enum import Enum

from typing import List


# HELPER FUNCTIONS


def split_at_first_whitespace(raw_text: str):

    # this may be that one case where regex is helpful

    p = -1  # pointer
    c = -1  # counter
    while p == -1 and c < len(raw_text) - 1:
        c += 1
        if raw_text[c] in string.whitespace:
            p = c

    if p == -1:
        # target without text
        return raw_text, ""
    else:
        # target and text
        return raw_text[:p], raw_text[p:].strip()


# CORE CLASSES


class LineType(Enum):
    BLANK = 0
    REGULAR = 1
    LINK = 2
    HEADING1 = 31
    HEADING2 = 32
    HEADING3 = 33
    LIST_ITEM = 4
    QUOTE = 5
    PREFORMAT_START = 61
    PREFORMAT_LINE = 62
    PREFORMAT_END = 63

    @staticmethod
    def identify(raw_text: str, preformat: bool):

        # note comments that explain when the order of tests matters

        # test preformatting states first and in order
        if raw_text.startswith("```") and preformat:
            return LineType.PREFORMAT_END

        elif raw_text.startswith("```") and not preformat:
            return LineType.PREFORMAT_START

        elif preformat:
            return LineType.PREFORMAT_LINE

        # test all other states after preformatting
        elif raw_text.strip() == "":
            return LineType.BLANK

        # test headings in descending order of length
        elif raw_text.startswith("###"):
            return LineType.HEADING3

        elif raw_text.startswith("##"):
            return LineType.HEADING2

        elif raw_text.startswith("#"):
            return LineType.HEADING1

        # simple startswith tests
        elif raw_text.startswith("*"):
            return LineType.LIST_ITEM

        elif raw_text.startswith("=>"):
            return LineType.LINK

        elif raw_text.startswith(">"):
            return LineType.QUOTE

        # all other cases are regular text (or not specified)
        else:
            return LineType.REGULAR

    @staticmethod
    def within_preformat(line_type):
        return line_type in [LineType.PREFORMAT_START, LineType.PREFORMAT_LINE]


@dataclass
class Line:
    type: LineType
    text: str
    extra: str

    @staticmethod
    def extract(line_type: LineType, raw_text: str):

        if line_type == LineType.BLANK:
            return Line(LineType.BLANK, "", "")

        elif line_type == LineType.REGULAR:
            return Line(LineType.REGULAR, raw_text.rstrip(), extra="")

        elif line_type == LineType.LINK:
            raw_text = raw_text[2:].strip()  # trim identifier
            target, clean_text = split_at_first_whitespace(raw_text)
            return Line(LineType.LINK, clean_text, extra=target)

        elif line_type == LineType.HEADING1:
            clean_text = raw_text[1:].strip()
            return Line(LineType.HEADING1, clean_text, extra="")

        elif line_type == LineType.HEADING2:
            clean_text = raw_text[2:].strip()
            return Line(LineType.HEADING2, clean_text, extra="")

        elif line_type == LineType.HEADING3:
            clean_text = raw_text[3:].strip()
            return Line(LineType.HEADING3, clean_text, extra="")

        elif line_type == LineType.LIST_ITEM:
            clean_text = raw_text[1:].strip()
            return Line(LineType.LIST_ITEM, clean_text, extra="")

        elif line_type == LineType.QUOTE:
            clean_text = raw_text[1:].strip()
            return Line(LineType.QUOTE, clean_text, extra="")

        elif line_type == LineType.PREFORMAT_START:
            label = raw_text[3:].strip()
            return Line(LineType.PREFORMAT_START, "", extra=label)

        elif line_type == LineType.PREFORMAT_LINE:
            return Line(LineType.PREFORMAT_LINE, raw_text.rstrip(), extra="")

        elif line_type == LineType.PREFORMAT_END:
            return Line(LineType.PREFORMAT_END, "", "")

        else:
            raise RuntimeWarning(f"Line type {line_type} not recognised.")


@dataclass
class Document:
    _lines: List[Line] = field(default_factory=lambda: [])
    _preformat: bool = False

    def append(self, raw_text: str):
        line_type = LineType.identify(raw_text, self._preformat)
        self._preformat = LineType.within_preformat(line_type)
        self._lines.append(Line.extract(line_type, raw_text))

    def emit_line_objects(self, auto_tidy=True):

        # test for empty list
        if len(self._lines) == 0:
            return []

        # test for invalid state
        trailing = LineType.within_preformat(self._lines[-1].type)
        if trailing and not auto_tidy:
            raise RuntimeWarning("Preformat block still open.")

        # yield all lines
        for line in self._lines:
            yield line

        # yield tidy line, if required
        if trailing and auto_tidy:
            yield Line(LineType.PREFORMAT_END, "", "")

    def emit_trim_gmi(self):

        for line in self.emit_line_objects(auto_tidy=True):

            if line.type == LineType.BLANK:
                yield ""

            elif line.type == LineType.REGULAR:
                yield line.text

            elif line.type == LineType.LINK:
                if line.text == "":
                    yield f"=> {line.extra}"
                else:
                    yield f"=> {line.extra} {line.text}"

            elif line.type == LineType.HEADING1:
                yield f"# {line.text}"

            elif line.type == LineType.HEADING2:
                yield f"## {line.text}"

            elif line.type == LineType.HEADING3:
                yield f"### {line.text}"

            elif line.type == LineType.LIST_ITEM:
                yield f"* {line.text}"

            elif line.type == LineType.QUOTE:
                yield f"> {line.text}"

            elif line.type == LineType.PREFORMAT_START:
                if line.extra == "":
                    yield "```"
                else:
                    yield f"```{line.extra}"

            elif line.type == LineType.PREFORMAT_LINE:
                yield line.text

            elif line.type == LineType.PREFORMAT_END:
                yield "```"

            else:
                raise RuntimeWarning(f"Line type {line.type} not recognised.")



# trimgmi
# Copyright 2022 David Seaward and contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import os.path
import sys
from pathlib import Path

from trimgmi.sample import reformat_file


def search_and_trim(base_folder: Path):

    for pattern in ["**/*.gmi", "**/*.gemini"]:
        for file_path in base_folder.glob(pattern):
            reformat_file("GMI", file_path, file_path)


def main():
    if len(sys.argv) != 2:
        raise RuntimeError("Too many/too few arguments.\nUse `trimgmi <path>`")

    base_path = Path(sys.argv[1])

    if os.path.isfile(base_path):
        reformat_file("gmi", base_path, base_path)
    elif os.path.isdir(base_path):
        search_and_trim(base_path)
    else:
        raise RuntimeError(f"Path {base_path} not recognised.")


if __name__ == "__main__":
    main()



# trimgmi
# Copyright 2022 David Seaward and contributors
# SPDX-License-Identifier: GPL-3.0-or-later
import html
from pathlib import Path

from trimgmi import Document as GmiDocument
from trimgmi import LineType as GmiLineType


def emit_naive_html(document: GmiDocument):

    # Out of scope:
    # * HTML character encoding
    # * HTML validation

    dangling_p = False
    dangling_link_list = False
    dangling_regular_list = False
    for line in document.emit_line_objects(auto_tidy=True):

        line.text = html.escape(line.text)
        line.extra = html.escape(line.extra)

        # paragraph tags
        if line.type == GmiLineType.REGULAR and not dangling_p:
            dangling_p = True
            yield "<p>"

        elif line.type == GmiLineType.REGULAR and dangling_p:
            yield "<br/>"

        elif line.type != GmiLineType.REGULAR and dangling_p:
            dangling_p = False
            yield "</p>"

        # link list tags
        if line.type == GmiLineType.LINK and not dangling_link_list:
            dangling_link_list = True
            yield "<ul>"

        elif line.type != GmiLineType.LINK and dangling_link_list:
            dangling_link_list = False
            yield "</ul>"

        # regular list tags
        if line.type == GmiLineType.LIST_ITEM and not dangling_regular_list:
            dangling_regular_list = True
            yield "<ul>"

        elif line.type != GmiLineType.LIST_ITEM and dangling_regular_list:
            dangling_regular_list = False
            yield "</ul>"

        # content

        if line.type == GmiLineType.REGULAR:
            yield f"{line.text}"

        elif line.type == GmiLineType.LINK:
            if line.text == "":
                yield f'<li><a href="{line.extra}">{line.extra}</a></li>'
            else:
                yield f'<li><a href="{line.extra}">{line.text}</a></li>'

        elif line.type == GmiLineType.HEADING1:
            yield f"<h1>{line.text}</h1>"

        elif line.type == GmiLineType.HEADING2:
            yield f"<h2>{line.text}</h2>"

        elif line.type == GmiLineType.HEADING3:
            yield f"<h3>{line.text}</h3>"

        elif line.type == GmiLineType.LIST_ITEM:
            yield f"<li>{line.text}</li>"

        elif line.type == GmiLineType.QUOTE:
            yield f"<blockquote>{line.text}</blockquote>"

        elif line.type == GmiLineType.PREFORMAT_START:
            if line.extra == "":
                yield "<pre>"
            else:
                yield f'<pre class="{line.extra}">'

        elif line.type == GmiLineType.PREFORMAT_LINE:
            yield line.text

        elif line.type == GmiLineType.PREFORMAT_END:
            yield "</pre>"

        elif line.type == GmiLineType.BLANK:
            pass  # whitespace is handled by HTML blocks

        else:
            raise RuntimeWarning(f"Line type {line.type} not recognised.")

    if dangling_p:
        yield "</p>"

    if dangling_link_list:
        yield "</ul>"

    if dangling_regular_list:
        yield "</ul>"


def emit_commonmark(document: GmiDocument):

    for line in document.emit_line_objects(auto_tidy=True):

        if line.type == GmiLineType.BLANK:
            yield ""

        elif line.type == GmiLineType.REGULAR:
            yield line.text

        elif line.type == GmiLineType.LINK:
            if line.text == "":
                yield f"* <{line.extra}>"
            else:
                yield f"* [{line.text}]({line.extra})"

        elif line.type == GmiLineType.HEADING1:
            yield f"# {line.text}"

        elif line.type == GmiLineType.HEADING2:
            yield f"## {line.text}"

        elif line.type == GmiLineType.HEADING3:
            yield f"### {line.text}"

        elif line.type == GmiLineType.LIST_ITEM:
            yield f"* {line.text}"

        elif line.type == GmiLineType.QUOTE:
            yield f"> {line.text}"

        elif line.type == GmiLineType.PREFORMAT_START:
            yield f"```{line.extra}"

        elif line.type == GmiLineType.PREFORMAT_LINE:
            yield line.text

        elif line.type == GmiLineType.PREFORMAT_END:
            yield "```"

        else:
            raise RuntimeWarning(f"Line type {line.type} not recognised.")


def reformat_file(output_format: str, input_path: Path, output_path: Path):

    # read input file
    document = GmiDocument()
    with open(input_path) as i:
        for line in i.readlines():
            document.append(line)

    # write trimmed gemtext
    if output_format == "GMI":
        with open(output_path, "w") as o:
            for line in document.emit_trim_gmi():
                o.write(line + "\n")

    # write naive HTML
    elif output_format == "HTML":
        with open(output_path, "w") as o:
            for line in emit_naive_html(document):
                o.write(line + "\n")

    # write naive HTML
    elif output_format == "MD":
        with open(output_path, "w") as o:
            for line in emit_commonmark(document):
                o.write(line + "\n")

    # alert on error
    else:
        raise RuntimeWarning("Output format not recognised.")



# trimgmi
# Copyright 2022 David Seaward and contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
from pathlib import Path

from trimgmi.sample import reformat_file


def main():
    if len(sys.argv) != 4:
        raise RuntimeError(
            "Too many/too few arguments.\n"
            "Use `convertgmi <format> <input path> <output path>`.\n"
            "Format can be gmi, md or html."
        )

    output_format = str.upper(sys.argv[1])
    input_path = Path(sys.argv[2])
    output_path = Path(sys.argv[3])
    reformat_file(output_format, input_path, output_path)


if __name__ == "__main__":
    main()
